require 'webrick/config'
require 'webrick/httpstatus'
require 'webrick/httpauth/authenticator'
require 'mutuallib'
  
module WEBrick
  module Config
    MutualAuth = {

    }
  end
  
  module HTTPAuth
    class MutualAuth
      include Authenticator

      AuthScheme = "Mutual"
      ResponseControlField = "Authentication-Control"

      HeaderIn = Struct.new(:rclass, :validation, :user, :kc1, :sid, :nc, :vkc)
      HeaderOut = Struct.new(:rclass, :authenticated, :validation, :sid, :ks1, :vks, :reason)
      Session = Struct.new(:sid, :realm, :kc1, :ks1, :z, :user, :nc_used, :nc_min, :time_out)

      NC_MAX = 0xFFFF
      NC_WND = 16
      TIMEOUT = 600
      SESSION_MAX = 32

      VERSION = "-wg-draft04"

      def self.algorithm_new(name)
        case name
        when "iso-kam3-dl-2048-sha256"
          return AuthMutual::DL2048.new
        when "iso-kam3-dl-4096-sha512"
          return AuthMutual::DL4096.new
        when "iso-kam3-ec-p256-sha256"
          return AuthMutual::EC256.new
        when "iso-kam3-ec-p521-sha512"
          return AuthMutual::EC521.new
        else
          raise ArgumentError.new('unsupported Algorithm')
        end
      end

      def self.make_passwd(conf, user, pass)
        a_name = conf[:Algorithm]
        a = algorithm_new(a_name)
        a.make_passwd(user, pass, a_name, conf[:Realm], conf[:AuthDomain])
      end

      def initialize(config, default = Config::MutualAuth)
        check_init(config)
        @resp_ctrl_field = self::class::ResponseControlField

        @config = default.dup.update(config)
        @mutex = Mutex.new
        @sessions = @config[:Sessions] || Hash.new
        @algorithm = @config[:Algorithm]
        @pwd_hash = @config[:PwdHash]
        @auth_domain = @config[:AuthDomain]
        @path = @config[:Path]
        @validation_host = @config[:ValidationHost]
        @optional_auth = @config[:OptionalAuth] ? true : false

        @a = MutualAuth.algorithm_new(@algorithm)

        unless @path
          raise ArgumentError.new('Path parameter is required')
        end
      end

      attr_reader :algorithm, :pwd_hash, :auth_domain, :path, :validation_host

      def authenticate(req, res)
        hdr_out = @mutex.synchronize { _authenticate(req, res) }
        unless hdr_out.authenticated
          challenge(req, res, hdr_out)
        end
        if hdr_out.rclass == :res_200_Optional_INIT
          res["Optional-WWW-Authenticate"] = auth_string(hdr_out)
        else
          res[@resp_info_field] = auth_string(hdr_out)
        end
        c = control_string(hdr_out)
        res[@resp_ctrl_field] = c if !c.empty?
        return true
      end

      def challenge(req, res, hdr_out = nil)
        hdr_out ||= default_header_out(req, res)
        res[@response_field] = auth_string(hdr_out)
        c = control_string(hdr_out)
        res[@resp_ctrl_field] = c if !c.empty?
        raise @auth_exception
      end

      private

      def auth_string(hdr_out)
        esc_w = esc_o = true
        esc_w = esc_o = true if @a.is_a?(AuthMutual::DL)
        esc_w = esc_o = false if @a.is_a?(AuthMutual::EC)
        param = Hash.new
        case hdr_out.rclass
        when :res_401_INIT
          param["reason"] = hdr_out.reason
          param["pwd-hash"] = @pwd_hash if @pwd_hash
        when :res_200_Optional_INIT
          param["pwd-hash"] = @pwd_hash if @pwd_hash
        when :res_401_KEX_S1
          param["sid"] = hdr_out.sid
          param["ks1"] = esc_w ? HTTPUtils::quote(hdr_out.ks1) : hdr_out.ks1
          param["nc-max"] = NC_MAX.to_s
          param["nc-window"] = NC_WND.to_s
          param["time"] = TIMEOUT
          param["path"] = HTTPUtils::quote(@path)
        when :res_200_VFY_S
          param["sid"] = hdr_out.sid
          param["vks"] = esc_o ? HTTPUtils::quote(hdr_out.vks) : hdr_out.vks

          s = "#{@auth_scheme} " + param.map{|k,v| "#{k}=#{v}"}.join(", ")
          info("resopnse res_200_B4: #{s}")
          return s
        end
        param["algorithm"] = @algorithm
        param["realm"] = HTTPUtils::quote(@realm)
        param["validation"] = hdr_out.validation
        param["auth-domain"] = HTTPUtils::quote(@auth_domain) if @auth_domain
        param["version"] = VERSION
        s = "#{@auth_scheme} " + param.map{|k,v| "#{k}=#{v}"}.join(", ")
        info("response #{hdr_out.rclass}: #{s}")
        return s
      end

      def control_string(hdr_out)
        param = Hash.new
        case hdr_out.rclass
        when :res_401_INIT
          param["location-when-unauthenticated"] = HTTPUtils::quote(@config[:LoginURL]) if @config[:LoginURL]
        when :res_200_VFY_S
          param["location-when-logout"] = HTTPUtils::quote(@config[:LogoutURL]) if @config[:LogoutURL]
        end
        return "" if param.empty?
        s = "#{@auth_scheme} " + param.map{|k,v| "#{k}=#{v}"}.join(", ")
        return s
      end

      def default_header_out(req, res)
        hdr_out = HeaderOut.new
        hdr_out.rclass = :res_401_INIT
        hdr_out.reason = "initial"
        hdr_out.authenticated = false
        hdr_out.validation = req.ssl? ? "tls-server-end-point" : "host"
        hdr_out
      end

      def _authenticate(req, res)
        hdr_out = default_header_out(req, res)
        unless mutual_credentials = check_scheme(req)
          if @optional_auth
            hdr_out.rclass = :res_200_Optional_INIT
            hdr_out.authenticated = true
          end
          return hdr_out
        end
        hdr_in = parse_header(mutual_credentials)
        STDERR.puts "from client to server: " + hdr_in.inspect
        return hdr_out unless hdr_in

        return hdr_out if hdr_in.validation.casecmp(hdr_out.validation) != 0

        case hdr_in.rclass
        when :req_KEX_C1
          info("request req_A1")
          verifier = @userdb.get_passwd(@realm, hdr_in.user, @reaload_db)
          if verifier
            user = hdr_in.user
            v  = @a.decodeGe(verifier)
          else
            user = nil
            v = @a.isDL ? 1 : @a.inf
          end
          kc1 = @a.decodeGe(hdr_in.kc1)
          return hdr_out unless @a.KAM3_T(kc1)
          s_s1 = nil
          ks1 = nil
          loop {
            if @a.isDL
              s_s1 = rand(@a.q)
            else
              s_s1 = rand(@a.r)
            end
            ks1 = @a.KAM3_C(s_s1, v, kc1)
            break if ks1
          }
          z = @a.KAM3_Vb(s_s1, kc1, ks1)
          rand32 = (0...4).map{ rand(256) }.pack("C*")
          sid = Digest::SHA1.hexdigest(rand32)
          s = Session.new(sid, @realm, kc1, ks1, z, user, 0, 0, Time.now)
          if @sessions.size >= SESSION_MAX
            arr = @sessions.map{|key, v| v}
            arr.sort! {|s1, s2| s2.time_out <=> s1.time_out}
            @sessions = {}
            arr[0, SESSION_MAX / 2].each {|v|
              @sessions[v.sid] = v
            }
          end
          @sessions[sid] = s
          hdr_out.rclass = :res_401_KEX_S1
          hdr_out.sid = sid
          hdr_out.ks1 = @a.encodeGe(ks1)
          return hdr_out
        when :req_VFY_C
          info("request req_A3")
          s = @sessions[hdr_in.sid]
          if !s
            hdr_out.rclass = :res_401_INIT
            hdr_out.reason = 'stale-session'
            return hdr_out
          elsif s.realm != @realm
            hdr_out.rclass = :res_401_INIT
            hdr_out.reason = 'stale-session'
            return hdr_out
          elsif !s.user
            hdr_out.rclass = :res_401_INIT
            hdr_out.reason = 'auth-failed'
            return hdr_out
          end
          n = Time.now
          if s.time_out + TIMEOUT < n
            @sessions.delete(hdr_in.sid)
            hdr_out.rclass = :res_401_INIT
            hdr_out.reason = 'reauth-neaded'
            return hdr_out
          end
          s.time_out = n

          if hdr_in.nc > NC_MAX
            hdr_out.rclass = :res_401_INIT
            hdr_out.reason = 'invalid-parameters'
            return hdr_out
          end
          if hdr_in.nc < s.nc_min
            hdr_out.rclass = :res_401_INIT
            hdr_out.reason = 'invalid-parameters'
            return hdr_out
          end
          nc_pos = hdr_in.nc - s.nc_min
          if nc_pos >= NC_WND
            shift = nc_pos - NC_WND + 1
            s.nc_used >>= shift
            s.nc_min += shift
            nc_pos = NC_WND - 1
          else
            if s.nc_used[nc_pos] == 1
              hdr_out.rclass = :res_401_INIT
              hdr_out.reason = 'invalid-parameters'
              return hdr_out
            end
          end
          s.nc_used |= 1 << nc_pos
          vh = @validation_host
          if hdr_out.validation.casecmp('tls-server-end-point') == 0
            vh = @a.tls_server_end_point(req.server_cert)
          end
          vkc = @a.hash_4("\4", s.kc1, s.ks1, s.z, hdr_in.nc, vh)
          vks = @a.hash_4("\3", s.kc1, s.ks1, s.z, hdr_in.nc, vh)

          unless vkc == @a.decodeHash(hdr_in.vkc)
            return hdr_out
          end
          req.user = s.user
          hdr_out.rclass = :res_200_VFY_S
          hdr_out.authenticated = true
          hdr_out.sid = hdr_in.sid
          hdr_out.vks = @a.encodeHash(vks)
          return hdr_out
        else
          return hdr_out
        end
      end

      def parse_header(header)
        args = {}

        n = 0
        while header =~ / *,? *([^=]+)=(\"[^\"]+\"|[^, ]+)/
          k = $1
          v = $2
          v = v[1, v.length - 2] if v[0,1] == '"'
          header = $'
          args[k] = v
          n += 1
        end

        return nil unless args["algorithm"].casecmp(@algorithm) == 0
        return nil unless args["realm"] == @realm
        
        hdr_in = HeaderIn.new
        hdr_in.validation = args["validation"]
        
        if args["algorithm"] && args["realm"] && args["validation"] && args["user"] && args["kc1"] && args["realm"]
          hdr_in.rclass = :req_KEX_C1
          hdr_in.user = args["user"]
          hdr_in.kc1 = args["kc1"]
          return hdr_in
        end
        if args["algorithm"] && args["realm"] && args["validation"] && args["sid"] && args["nc"] && args["vkc"] && args["nc"] =~ %r!\A(0|[1-9][0-9]*)\z!
          hdr_in.rclass = :req_VFY_C
          hdr_in.sid = args["sid"]
          hdr_in.nc = args["nc"].to_i
          hdr_in.vkc = args["vkc"]
          return hdr_in
        end
        return nil
      end
    end
  end
end
