require 'digest/sha1'
require 'digest/sha2'
require 'base64'
require 'openssl'

module AuthMutual
  class Algorithm
    def initialize(is_dl, hash_cls, hash_len, hmac_func)
      @isDL = is_dl
      @hash_cls = hash_cls
      @hash_len = hash_len
      @hmac_func = hmac_func
    end

    attr_reader :isDL, :hash_cls, :hash_len, :hmac_func

    def hexhash(octs)
      hash_cls.new.update(octs || "").hexdigest
    end

    def octets(bn, byte_len)
      [sprintf('%0*x',byte_len*2, bn)].pack("H*")
    end

    def hashoctets(s)
      octets(hexhash(s).hex, hash_len/8)
    end

    def tls_server_end_point(cert)
      alg = case cert.signature_algorithm
            when /sha1/i,/md5/i
              "sha256"
            when /sha224/ii,/sha256/i,/sha512/i,/sha384/i,/md4/i,/md2/i,/mdc2/i,/ripemd160/i
              $&.downcase
            else
              "sha256"
            end
      OpenSSL::Digest.new(alg).digest(cert.to_der)
    end

    def make_passwd(user, pass, alg_name, realm, auth_domain)
      salt = vs(alg_name) + vs(auth_domain) + vs(realm) + vs(user)
      pi = OpenSSL::PKCS5.pbkdf2_hmac(pass, salt, 16384,
                                      hash_len / 8,
                                      hmac_func)

      # convert binary sequense (null-terminated String) to big integer
      pi = pi.unpack("H*").first.hex
      printf "pi = %#x\n", pi
      encodeGe(KAM3_J(pi))
    end

    def KAM3_J(pi)
      if isDL
        powmod(g, pi, q)
      else
        g * pi
      end
    end

    def KAM3_D(sa)
      if isDL
        powmod(g, sa, q)
      else
        g * sa
      end
    end

    def KAM3_T(wa)
      if isDL
        return false if wa <= 1
        return false if wa >= q-1
      else
        y = wa * cof_2
        return false if y.x == nil
      end
      true
    end

    def KAM3_C(sb, v, wa)
      e = hexhash("\1" + octetsGe(wa)).hex
      if isDL
        x = powmod(v * powmod(wa, e, q), sb, q)
        return nil if x == 1
        return nil if x == q-1
        return x
      else
        x = (v + wa * e) * sb
        y = x * 4
        return nil if y.x == nil
        return x
      end
    end

    def KAM3_Va(pi, sa, wa, wb, h_1, h_2)
      if isDL
        powmod(wb, div(sa + h_2, sa * h_1 + pi, r), q)
      else
        wb * div(sa + h_2, sa * h_1 + pi, r)
      end
    end

    def KAM3_Vb(sb, wa, wb)
      d = hexhash("\2" + octetsGe(wa) + octetsGe(wb)).hex
      if isDL
        return powmod(wa * powmod(g, d, q), sb, q)
      else
        return (wa + g * d) * sb
      end
    end

    def vi(n)
      n = n.to_i
      [n].pack('w')
    end

    def vs(s)
      s = s.to_s
      [s.length].pack('w') + s
    end

    def hash_4(hdr, wa, wb, z, nc, v)
      hexhash(hdr + octetsGe(wa) + octetsGe(wb) + octetsGe(z) + vi(nc) + vs(v)).hex
    end

    def powmod(x, y, z)
      # compute x ** y % z (semi) efficiently
      v = 1
      x = x % z
      while y > 0
        v = v * x % z if y & 1 == 1
        x = x * x % z
        y >>= 1
      end
      v
    end

    def inv(v, r)
      a, b = v, r
      x = 0
      lastx = 1
      y = 1
      lasty = 0
      while (b != 0)
        quo, b, a = a / b, a % b, b #/;
        x, lastx = (lastx - quo * x + r) % r, x
        y, lasty = (lasty - quo * y + r) % r, y
      end
      raise 'no inverse element exists in inv()' if a != 1
      return lastx
    end

    def div(x, s, r)
      x * inv(s, r) % r
    end
  end

  class DL < Algorithm
    def initialize(q, g, r, bits, hash_cls, hash_len, hmac_func)
      super(true, hash_cls, hash_len, hmac_func)
      @q = q
      @g = g
      @r = r
      @bits = bits
    end

    attr_reader :q, :g, :r, :bits

    def octetsGe(s)
      octets(s, bits/8)
    end

    def encodeGe(s)
      Base64.encode64(octetsGe(s)).gsub("\n", "")
    end

    def decodeGe(s)
      raise if s.length != (bits / 8 + 2) / 3 * 4
      Base64.decode64(s).unpack("H*")[0].hex;
    end

    def encodeHash(s)
      Base64.encode64(octets(s, hash_len / 8)).gsub("\n", "")
    end

    def decodeHash(s)
      raise if s.length != (hash_len / 8 + 2) / 3 * 4
      Base64.decode64(s).unpack("H*")[0].hex;
    end
  end

  class DL2048 < DL
    G = 2
    Q_s = %q(
      FFFFFFFF FFFFFFFF C90FDAA2 2168C234 C4C6628B 80DC1CD1
      29024E08 8A67CC74 020BBEA6 3B139B22 514A0879 8E3404DD
      EF9519B3 CD3A431B 302B0A6D F25F1437 4FE1356D 6D51C245
      E485B576 625E7EC6 F44C42E9 A637ED6B 0BFF5CB6 F406B7ED
      EE386BFB 5A899FA5 AE9F2411 7C4B1FE6 49286651 ECE45B3D
      C2007CB8 A163BF05 98DA4836 1C55D39A 69163FA8 FD24CF5F
      83655D23 DCA3AD96 1C62F356 208552BB 9ED52907 7096966D
      670C354E 4ABC9804 F1746C08 CA18217C 32905E46 2E36CE3B
      E39E772C 180E8603 9B2783A2 EC07A28F B5C55DF0 6F4C52C9
      DE2BCBF6 95581718 3995497C EA956AE5 15D22618 98FA0510
      15728E5A 8AACAA68 FFFFFFFF FFFFFFFF
    ).gsub(/\s/, "")
    Q = Q_s.hex
    R = Q >> 1
    def initialize
      super(Q, G, R, 2048, Digest::SHA256, 256, "sha256")
    end
  end

  class DL4096 < DL
    G = 2
    Q_s =%q(
      FFFFFFFF FFFFFFFF C90FDAA2 2168C234 C4C6628B 80DC1CD1
      29024E08 8A67CC74 020BBEA6 3B139B22 514A0879 8E3404DD
      EF9519B3 CD3A431B 302B0A6D F25F1437 4FE1356D 6D51C245
      E485B576 625E7EC6 F44C42E9 A637ED6B 0BFF5CB6 F406B7ED
      EE386BFB 5A899FA5 AE9F2411 7C4B1FE6 49286651 ECE45B3D
      C2007CB8 A163BF05 98DA4836 1C55D39A 69163FA8 FD24CF5F
      83655D23 DCA3AD96 1C62F356 208552BB 9ED52907 7096966D
      670C354E 4ABC9804 F1746C08 CA18217C 32905E46 2E36CE3B
      E39E772C 180E8603 9B2783A2 EC07A28F B5C55DF0 6F4C52C9
      DE2BCBF6 95581718 3995497C EA956AE5 15D22618 98FA0510
      15728E5A 8AAAC42D AD33170D 04507A33 A85521AB DF1CBA64
      ECFB8504 58DBEF0A 8AEA7157 5D060C7D B3970F85 A6E1E4C7
      ABF5AE8C DB0933D7 1E8C94E0 4A25619D CEE3D226 1AD2EE6B
      F12FFA06 D98A0864 D8760273 3EC86A64 521F2B18 177B200C
      BBE11757 7A615D6C 770988C0 BAD946E2 08E24FA0 74E5AB31
      43DB5BFC E0FD108E 4B82D120 A9210801 1A723C12 A787E6D7
      88719A10 BDBA5B26 99C32718 6AF4E23C 1A946834 B6150BDA
      2583E9CA 2AD44CE8 DBBBC2DB 04DE8EF9 2E8EFC14 1FBECAA6
      287C5947 4E6BC05D 99B2964F A090C3A2 233BA186 515BE7ED
      1F612970 CEE2D7AF B81BDD76 2170481C D0069127 D5B05AA9
      93B4EA98 8D8FDDC1 86FFB7DC 90A6C08F 4DF435C9 34063199
      FFFFFFFF FFFFFFFF
    ).gsub(/\s/, "")
    Q = Q_s.hex
    R = Q >> 1
    def initialize
      super(Q, G, R, 4096, Digest::SHA512, 512, "sha512")
    end
  end

  class ECPoint
    def initialize(x, y, ec)
      @x = x
      @y = y
      @ec = ec
      if x
        if (y*y-(x*x+ec.a)*x-ec.b) % ec.p != 0
          raise
        end
      end
    end
    attr_reader :x, :y, :ec

    def +(q)
      p = self
      return q if p.x == nil
      return p if q.x == nil
      if p.x == q.x
        if p.y == q.y
          if p.y == 0
            return ec.inf
          else
            s1 = 3 * p.x * p.x + ec.a
            s2 = ec.inv(2 * p.y, ec.p)
            s = s1 * s2 % ec.p
            nx = (s*s - p.x - p.x) % ec.p
            ny = (s * (p.x - nx) - p.y) % ec.p
            return ECPoint.new(nx, ny, ec)
          end
        else # p.y != q.y
          return ec.inf
        end
      else # p.x != q.x
        s1 = p.y - q.y
        s2 = ec.inv(p.x - q.x + ec.p, ec.p)
        s = s1 * s2 % ec.p
        nx = (s*s - p.x - q.x) % ec.p
        ny = (s * (p.x - nx) - p.y) % ec.p
        return ECPoint.new(nx, ny, ec)
      end
    end

    def *(n)
      n = n % ec.r

      v = ec.inf
      x = self
      while n > 0
        v = v + x if n[0] == 1
        x = x + x
        n >>= 1
      end
      v
    end

    def to_s(radix = 10)
      if x
        "[#{x.to_s(radix)},#{y.to_s(radix)}]"
      else
        "INF"
      end
    end
  end

  class EC < Algorithm
    def initialize(p, r, a, b, gx, gy, cof_2, cof_p, bits, hash_cls, hash_len, hmac_func)
      super(false, hash_cls, hash_len, hmac_func)
      @p = p
      @r = r
      @a = a
      @b = b
      @cof_2 = cof_2
      @cof_p = cof_p
      @bits = bits
      @g = ECPoint.new(gx, gy, self)
      @inf = ECPoint.new(nil, nil, self)
    end

    attr_reader :p, :r, :a, :b, :cof_2, :cof_p, :g, :inf, :bits

    def octetsGe(s)
      octets(2 * s.x + s.y[0], (bits + 1 + 7) / 8)
    end

    def encodeGe(s)
      octetsGe(s).unpack("H*")[0]
    end

    def decodeGe(s)
      raise if s.length != (bits + 1 + 7) / 4
      v = s.hex
      x = v >> 1
      raise if p % 4 != 3
      t = ((x * x + a) * x + b) % p
      u = p >> 2
      y = powmod(t, u+1, p)
      z = y * y % p
      raise if z != t
      if y[0] != v[0]
        y = p - y
      end
      ECPoint.new(x, y, self)
    end

    def encodeHash(s)
      octets(s, hash_len / 8).unpack("H*")[0]
    end

    def decodeHash(s)
      raise if s.length != (hash_len / 4)
      s.hex
    end
  end

  class EC256 < EC
    def initialize
      super(
        115792089210356248762697446949407573530086143415290314195533631308867097853951,
        115792089210356248762697446949407573529996955224135760342422259061068512044369,
        115792089210356248762697446949407573530086143415290314195533631308867097853948,
        0x5ac635d8aa3a93e7b3ebbd55769886bc651d06b0cc53b0f63bce3c3e27d2604b,
        0x6b17d1f2e12c4247f8bce6e563a440f277037d812deb33a0f4a13945d898c296,
        0x4fe342e2fe1a7f9b8ee7eb4a7c0f9e162bce33576b315ececbb6406837bf51f5,
        1,
        1,
        256,
        Digest::SHA256, 256, "sha256"
      )
    end
  end

  class EC521 < EC
    def initialize
      super(
        6864797660130609714981900799081393217269435300143305409394463459185543183397656052122559640661454554977296311391480858037121987999716643812574028291115057151,
        6864797660130609714981900799081393217269435300143305409394463459185543183397655394245057746333217197532963996371363321113864768612440380340372808892707005449,
        6864797660130609714981900799081393217269435300143305409394463459185543183397656052122559640661454554977296311391480858037121987999716643812574028291115057148,
        0x051953eb9618e1c9a1f929a21a0b68540eea2da725b99b315f3b8b489918ef109e156193951ec7e937b1652c0bd3bb1bf073573df883d2c34f1ef451fd46b503f00,
        0xc6858e06b70404e9cd9e3ecb662395b4429c648139053fb521f828af606b4d3dbaa14b5e77efe75928fe1dc127a2ffa8de3348b3c1856a429bf97e7e31c2e5bd66,
        0x11839296a789a3bc0045c8a5fb42c7d1bd998f54449579b446817afbd17273e662c97ee72995ef42640c550b9013fad0761353c7086a272c24088be94769fd16650,
        1,
        1,
        521,
        Digest::SHA512, 512, "sha512"
      )
    end
  end
end
