require 'yaml'
require 'webrick'
require 'webrick/https'
require 'erb'
$: << File.dirname(__FILE__) + '/..'
require 'mutual'
require 'htmutual'
if RUBY_VERSION > "1.9"
  require_relative 'brokenauth.rb'
else
  # 1.8.* or older
  require 'brokenauth'
end
include WEBrick

$conf_yaml = ARGV[0]
unless $conf_yaml
  puts "usage: ruby main.rb server-conf.yml"
  exit 1
end

module WEBrick
  class HTTPResponse
    include ERB::Util
    #
    # Custom Error page
    #
    def create_error_page
      @header['content-type'] = "text/plain"
      @body =  "Error #{@reason_phrase}\n"
      auth = authenticator
      if auth
        if BrokenAuth::Base === auth
          auth = authenticator.auth
        end
        if HTTPAuth::MutualAuth === auth
          if $conf['UnauthPage']
            @header['content-type'] = "text/html; charset=UTF-8"
            lang = /^ja/ =~ request['Accept-Language'] ? 'ja' : 'en'
            @body = ERB.new(IO.read($conf['UnauthPage'][lang])).result(binding)
          else
            @body << "Configuration of Mutual Auth\n"
            @body << "  Realm: #{auth.realm}\n"
            @body << "  Algorithm: #{auth.algorithm}\n"
            @body << "  PwdHash: #{auth.pwd_hash}\n"
            @body << "  AuthDomain: #{auth.auth_domain}\n"
            @body << "  Path: #{auth.path}\n"
            @body << "  ValidationHost: #{auth.validation_host}\n"
            @body << "  Header: #{@header["www-authenticate"]}"
          end
        end
      end
    end

    attr_accessor :authenticator
    attr_accessor :request
  end
end

$conf = YAML.load_file($conf_yaml)

$servers = []
$servers << HTTPServer.new(:Port => $conf["Port"])
if $conf["SSLEnable"]
  $servers << HTTPServer.new(
    :Port => $conf["SSLPort"],
    :SSLEnable => true,
    :SSLVerifyClient => ::OpenSSL::SSL::VERIFY_NONE,
    :SSLCertificate =>
    open($conf["SSLCertificate"], "r") {|io|
      OpenSSL::X509::Certificate.new(io)
    },
    :SSLPrivateKey =>
    open($conf["SSLPrivateKey"], "r") {|io|
      OpenSSL::PKey::RSA.new(io)
    },
    :SSLCACertificateFile => $conf["SSLCACertificate"]
  )
end

def logger
  $servers[0].logger
end

$userdb_list = {}
$default_pwfile = $conf["DefaultPasswordFile"]
$userdb_list[$default_pwfile] =  HTTPAuth::Htmutual.new($default_pwfile)
$auth_list = {}
$conf_list = {}

def normalize_validation_host(vh)
  def validation_host_is_default_port?(vh)
    ((/^https:.*:443$/ =~ vh) or (/^http:.*:80$/ =~ vh)) ? true : false
  end

  rv = vh.downcase
  if validation_host_is_default_port?(vh)
    rv = vh.sub(/:\d*$/, '')
  end
  return rv
end

def add_mutual_realm(userdb, key, users, realm, alg, vh, auth_domain, path, opt = {})
  o = opt.dup.update({
    :Realm => realm,
    :UserDB => userdb,
    :Algorithm => alg,
    :Path => path,
    :ValidationHost => normalize_validation_host(vh),
    :AuthDomain => auth_domain
  })
  users.each {|up|
    if up =~ /^(.*):(.*)$/
      user = $1
      pass = $2
      unless userdb.get_passwd(realm, user, false)
        userdb.set_passwd(o, user, pass)
        userdb.flush
      end
    end
  }
  auth = HTTPAuth::MutualAuth.new(o)
  $auth_list[key] = auth
  $conf_list[key] = o
end

$default_vh = $conf["DefaultValidationHost"]
$default_dom = $conf["DefaultAuthDomain"]
$default_users = $conf["DefaultUsers"] || ["user_1:passwd_1", "user_2:passwd_2", "user_3:passwd_3"]
$realm_list = $conf["RealmList"] || []
$realm_files = $conf["RealmFiles"] || []
$realm_files.each {|f|
  if File.exist?(f)
    r = YAML.load_file(f) || []
    $realm_list += r
  end
}
$session_table = Hash.new

logger.info("Generating password list ...")

$realm_list.each {|c|
  realm = c["Realm"]
  realm_ind = c["RealmIndex"] || realm
  vh = c["ValidationHost"] || $default_vh
  dm = c["AuthDomain"] || $default_dom
  pwfile = c["PasswordFile"] || $default_pwfile
  users = c["Users"] || $default_users
  if c["LoginURL"]
    login_url = c["LoginURL"]
  elsif c["LoginPath"]
    login_url = vh + c["LoginPath"]
  else
    login_url = nil
  end
  if c["LogoutURL"]
    logout_url = c["LogoutURL"]
  elsif c["LogoutPath"]
    logout_url = vh + c["LogoutPath"]
  else
    logout_url = nil
  end
  $userdb_list[pwfile] ||= HTTPAuth::Htmutual.new(pwfile)
  userdb = $userdb_list[pwfile]
  logger.info("  Add realm #{realm}")
  add_mutual_realm(userdb, realm_ind, users, realm, c["Algorithm"], vh, dm, c["Path"],
    {:Sessions => $session_table, :LoginURL => login_url, :LogoutURL => logout_url})
}

logger.info("Done.")

logger.info("Scanning 'htdoc/' ...")

$dirs = {}
Dir.glob('htdoc/**/*.yml').each {|yml|
  conf = YAML.load_file(yml)
  if conf
    dirname, fname = File.split(yml)
    if fname != 'authmutual.yml'
      dirname = yml.sub(/\.yml$/, '')
    end
    auth = nil
    realm = conf["Realm"]
    if realm
      logger.info("  Enable mutual authentication on '#{dirname}'")
      auth = $auth_list[realm]
      if auth
        cls_name = conf["BrokenAuthClass"]
        if cls_name
          auth = wrap_broken_auth(auth, cls_name, conf["BrokenAuthConf"], logger)
          logger.info("    Simulate broken mutual authentication on '#{dirname}'")
        end
        $dirs[dirname.sub(/^htdoc/, '')] = auth
      else
        logger.warn("  Realm '#{realm}' is not defined")
      end
    end
  end
}
$dirs['/'] ||= nil

logger.info("Done.")


$servers.each {|s|
  $dirs.each {|d, auth|
    file_handler = HTTPServlet::FileHandler.new(s, 'htdoc' + d, {:FancyIndexing => true, :UserDir => nil})
    s.mount_proc(d) {|req, res|
      res.authenticator = auth
      res.request = req
      auth.authenticate(req, res) if auth
      file_handler.service(req, res)
    }
  }
  s.mount_proc('/reset') {|req, res|
    reset_broken_auth
    res.set_redirect(HTTPStatus::SeeOther, '/')
  }
}

Signal.trap(:INT) { $servers.each {|s| s.shutdown }}
Signal.trap(:TERM) { $servers.each {|s| s.shutdown }}

$servers.map {|s|
  Thread.new {
    s.start
  }
}.each {|t|
  t.join
}
